@@ -29,13 +29,16 @@
 #include "latch.h"
 #include "openvswitch/hmap.h"
 #include "openvswitch/poll-loop.h"
+#include "openvswitch/types.h"
 #include "openvswitch/vlog.h"
 #include "ovs-atomic.h"
 #include "packets.h"
 #include "util.h"
 
 VLOG_DEFINE_THIS_MODULE(ipf);
+COVERAGE_DEFINE(ipf_stuck_frag_list_expired);
 COVERAGE_DEFINE(ipf_stuck_frag_list_purged);
+COVERAGE_DEFINE(ipf_l3csum_checked);
 COVERAGE_DEFINE(ipf_l3csum_err);
 
 enum {
@@ -77,7 +80,7 @@ enum {
 enum ipf_counter_type {
     IPF_NFRAGS_ACCEPTED,
     IPF_NFRAGS_COMPL_SENT,
-    IPF_NFRAGS_EXPD_SENT,
+    IPF_NFRAGS_EXPIRED,
     IPF_NFRAGS_TOO_SMALL,
     IPF_NFRAGS_OVERLAP,
     IPF_NFRAGS_PURGED,
@@ -434,7 +437,9 @@ ipf_reassemble_v4_frags(struct ipf_list *ipf_list)
     len += rest_len;
     l3 = dp_packet_l3(pkt);
     ovs_be16 new_ip_frag_off = l3->ip_frag_off & ~htons(IP_MORE_FRAGMENTS);
-    if (!dp_packet_hwol_is_ipv4(pkt)) {
+    if (dp_packet_hwol_tx_ip_csum(pkt)) {
+        dp_packet_ol_reset_ip_csum_good(pkt);
+    } else {
         l3->ip_csum = recalc_csum16(l3->ip_csum, l3->ip_frag_off,
                                     new_ip_frag_off);
         l3->ip_csum = recalc_csum16(l3->ip_csum, l3->ip_tot_len, htons(len));
@@ -487,9 +492,9 @@ ipf_reassemble_v6_frags(struct ipf_list *ipf_list)
     const void *data = l3 + 1;
     size_t datasize = pl;
 
-    const struct ovs_16aligned_ip6_frag *frag_hdr = NULL;
-    if (!parse_ipv6_ext_hdrs(&data, &datasize, &nw_proto, &nw_frag, &frag_hdr)
-        || !nw_frag || !frag_hdr) {
+    const struct ovs_16aligned_ip6_frag *frag_hdr;
+    if (!parse_ipv6_ext_hdrs(&data, &datasize, &nw_proto, &nw_frag, &frag_hdr,
+                             NULL) || !nw_frag || !frag_hdr) {
 
         ipf_print_reass_packet("Unparsed reassembled v6 packet; v6 hdr:", l3);
         dp_packet_delete(pkt);
@@ -581,11 +586,6 @@ ipf_list_state_transition(struct ipf *ipf, struct ipf_list *ipf_list,
 static bool
 ipf_is_valid_v4_frag(struct ipf *ipf, struct dp_packet *pkt)
 {
-    if (OVS_UNLIKELY(dp_packet_ip_checksum_bad(pkt))) {
-        COVERAGE_INC(ipf_l3csum_err);
-        goto invalid_pkt;
-    }
-
     const struct eth_header *l2 = dp_packet_eth(pkt);
     const struct ip_header *l3 = dp_packet_l3(pkt);
 
@@ -615,9 +615,12 @@ ipf_is_valid_v4_frag(struct ipf *ipf, struct dp_packet *pkt)
         goto invalid_pkt;
     }
 
-    if (OVS_UNLIKELY(!dp_packet_ip_checksum_valid(pkt)
-                     && !dp_packet_hwol_is_ipv4(pkt)
-                     && csum(l3, ip_hdr_len) != 0)) {
+    bool bad_csum = dp_packet_ip_checksum_bad(pkt);
+    if (OVS_UNLIKELY(!bad_csum && !dp_packet_ip_checksum_good(pkt))) {
+        COVERAGE_INC(ipf_l3csum_checked);
+        bad_csum = csum(l3, ip_hdr_len);
+    }
+    if (OVS_UNLIKELY(bad_csum)) {
         COVERAGE_INC(ipf_l3csum_err);
         goto invalid_pkt;
     }
@@ -685,9 +688,9 @@ ipf_is_valid_v6_frag(struct ipf *ipf, struct dp_packet *pkt)
     uint8_t nw_proto = l3->ip6_nxt;
     const void *data = l3 + 1;
     size_t datasize = l3_size - l3_hdr_size;
-    const struct ovs_16aligned_ip6_frag *frag_hdr = NULL;
+    const struct ovs_16aligned_ip6_frag *frag_hdr;
     if (!parse_ipv6_ext_hdrs(&data, &datasize, &nw_proto, &nw_frag,
-                             &frag_hdr) || !nw_frag || !frag_hdr) {
+                             &frag_hdr, NULL) || !nw_frag || !frag_hdr) {
         return false;
     }
 
@@ -728,9 +731,10 @@ ipf_v6_key_extract(struct dp_packet *pkt, ovs_be16 dl_type, uint16_t zone,
     uint8_t nw_proto = l3->ip6_nxt;
     const void *data = l3 + 1;
     size_t datasize = dp_packet_l3_size(pkt) - sizeof *l3;
-    const struct ovs_16aligned_ip6_frag *frag_hdr = NULL;
+    const struct ovs_16aligned_ip6_frag *frag_hdr;
 
-    parse_ipv6_ext_hdrs(&data, &datasize, &nw_proto, &nw_frag, &frag_hdr);
+    parse_ipv6_ext_hdrs(&data, &datasize, &nw_proto, &nw_frag, &frag_hdr,
+                        NULL);
     ovs_assert(nw_frag && frag_hdr);
     ovs_be16 ip6f_offlg = frag_hdr->ip6f_offlg;
     *start_data_byte = ntohs(ip6f_offlg & IP6F_OFF_MASK) +
@@ -1029,8 +1033,7 @@ ipf_purge_list_check(struct ipf *ipf, struct ipf_list *ipf_list,
  * with 'ipf_send_completed_frags()' and 'ipf_send_expired_frags()'. */
 static bool
 ipf_send_frags_in_list(struct ipf *ipf, struct ipf_list *ipf_list,
-                       struct dp_packet_batch *pb,
-                       enum ipf_list_type list_type, bool v6, long long now)
+                       struct dp_packet_batch *pb, bool v6, long long now)
     OVS_REQUIRES(ipf->ipf_lock)
 {
     if (ipf_purge_list_check(ipf, ipf_list, now)) {
@@ -1044,12 +1047,7 @@ ipf_send_frags_in_list(struct ipf *ipf, struct ipf_list *ipf_list,
             ipf_list->last_sent_idx++;
             atomic_count_dec(&ipf->nfrag);
 
-            if (list_type == IPF_FRAG_COMPLETED_LIST) {
-                ipf_count(ipf, v6, IPF_NFRAGS_COMPL_SENT);
-            } else {
-                ipf_count(ipf, v6, IPF_NFRAGS_EXPD_SENT);
-                pkt->md.ct_state = CS_INVALID;
-            }
+            ipf_count(ipf, v6, IPF_NFRAGS_COMPL_SENT);
 
             if (ipf_list->last_sent_idx == ipf_list->last_inuse_idx) {
                 return true;
@@ -1066,7 +1064,8 @@ ipf_send_frags_in_list(struct ipf *ipf, struct ipf_list *ipf_list,
  * cleans up the list context when it is empty.*/
 static void
 ipf_send_completed_frags(struct ipf *ipf, struct dp_packet_batch *pb,
-                         long long now, bool v6)
+                         long long now, bool v6, uint16_t zone,
+                         odp_port_t in_port)
 {
     if (ovs_list_is_empty(&ipf->frag_complete_list)) {
         return;
@@ -1076,11 +1075,22 @@ ipf_send_completed_frags(struct ipf *ipf, struct dp_packet_batch *pb,
     struct ipf_list *ipf_list;
 
     LIST_FOR_EACH_SAFE (ipf_list, list_node, &ipf->frag_complete_list) {
+
         if ((ipf_list->key.dl_type == htons(ETH_TYPE_IPV6)) != v6) {
             continue;
         }
-        if (ipf_send_frags_in_list(ipf, ipf_list, pb, IPF_FRAG_COMPLETED_LIST,
-                                   v6, now)) {
+        if (ipf_list->key.zone != zone) {
+            continue;
+        }
+
+        /* Check that the batch's in_port matches. */
+        struct dp_packet *pkt
+            = ipf_list->frag_list[ipf_list->last_sent_idx + 1].pkt;
+        if (in_port != pkt->md.in_port.odp_port) {
+            continue;
+        }
+
+        if (ipf_send_frags_in_list(ipf, ipf_list, pb, v6, now)) {
             ipf_completed_list_clean(&ipf->frag_lists, ipf_list);
         } else {
             break;
@@ -1090,12 +1100,9 @@ ipf_send_completed_frags(struct ipf *ipf, struct dp_packet_batch *pb,
     ovs_mutex_unlock(&ipf->ipf_lock);
 }
 
-/* Conservatively adds fragments associated with a expired fragment list to
- * a packet batch to be processed by the calling application, typically
- * conntrack. Also cleans up the list context when it is empty.*/
+/* Remove expired fragment lists and clean up the list context. */
 static void
-ipf_send_expired_frags(struct ipf *ipf, struct dp_packet_batch *pb,
-                       long long now, bool v6)
+ipf_delete_expired_frags(struct ipf *ipf, long long now)
 {
     enum {
         /* Very conservative, due to DOS probability. */
@@ -1112,21 +1119,23 @@ ipf_send_expired_frags(struct ipf *ipf, struct dp_packet_batch *pb,
     size_t lists_removed = 0;
 
     LIST_FOR_EACH_SAFE (ipf_list, list_node, &ipf->frag_exp_list) {
-        if ((ipf_list->key.dl_type == htons(ETH_TYPE_IPV6)) != v6) {
-            continue;
-        }
         if (now <= ipf_list->expiration ||
             lists_removed >= IPF_FRAG_LIST_MAX_EXPIRED) {
             break;
         }
 
-        if (ipf_send_frags_in_list(ipf, ipf_list, pb, IPF_FRAG_EXPIRY_LIST,
-                                   v6, now)) {
-            ipf_expiry_list_clean(&ipf->frag_lists, ipf_list);
-            lists_removed++;
-        } else {
-            break;
+        while (ipf_list->last_sent_idx < ipf_list->last_inuse_idx) {
+            struct dp_packet * pkt
+                = ipf_list->frag_list[ipf_list->last_sent_idx + 1].pkt;
+            dp_packet_delete(pkt);
+            atomic_count_dec(&ipf->nfrag);
+            COVERAGE_INC(ipf_stuck_frag_list_expired);
+            ipf_count(ipf, ipf_list->key.dl_type == htons(ETH_TYPE_IPV6),
+                      IPF_NFRAGS_EXPIRED);
+            ipf_list->last_sent_idx++;
         }
+        ipf_expiry_list_clean(&ipf->frag_lists, ipf_list);
+        lists_removed++;
     }
 
     ovs_mutex_unlock(&ipf->ipf_lock);
@@ -1208,7 +1217,9 @@ ipf_post_execute_reass_pkts(struct ipf *ipf,
                     } else {
                         struct ip_header *l3_frag = dp_packet_l3(frag_i->pkt);
                         struct ip_header *l3_reass = dp_packet_l3(pkt);
-                        if (!dp_packet_hwol_is_ipv4(frag_i->pkt)) {
+                        if (dp_packet_hwol_tx_ip_csum(frag_i->pkt)) {
+                            dp_packet_ol_reset_ip_csum_good(frag_i->pkt);
+                        } else {
                             ovs_be32 reass_ip =
                                 get_16aligned_be32(&l3_reass->ip_src);
                             ovs_be32 frag_ip =
@@ -1266,13 +1277,14 @@ ipf_preprocess_conntrack(struct ipf *ipf, struct dp_packet_batch *pb,
  * with low priority.  Reassembled packets are freed. */
 void
 ipf_postprocess_conntrack(struct ipf *ipf, struct dp_packet_batch *pb,
-                          long long now, ovs_be16 dl_type)
+                          long long now, ovs_be16 dl_type, uint16_t zone,
+                          odp_port_t in_port)
 {
     if (ipf_get_enabled(ipf) || atomic_count_get(&ipf->nfrag)) {
         bool v6 = dl_type == htons(ETH_TYPE_IPV6);
         ipf_post_execute_reass_pkts(ipf, pb, v6);
-        ipf_send_completed_frags(ipf, pb, now, v6);
-        ipf_send_expired_frags(ipf, pb, now, v6);
+        ipf_send_completed_frags(ipf, pb, now, v6, zone, in_port);
+        ipf_delete_expired_frags(ipf, now);
     }
 }
 
@@ -1445,7 +1457,7 @@ ipf_get_status(struct ipf *ipf, struct ipf_status *ipf_status)
                         &ipf_status->v4.nfrag_accepted);
     atomic_read_relaxed(&ipf->n4frag_cnt[IPF_NFRAGS_COMPL_SENT],
                         &ipf_status->v4.nfrag_completed_sent);
-    atomic_read_relaxed(&ipf->n4frag_cnt[IPF_NFRAGS_EXPD_SENT],
+    atomic_read_relaxed(&ipf->n4frag_cnt[IPF_NFRAGS_EXPIRED],
                         &ipf_status->v4.nfrag_expired_sent);
     atomic_read_relaxed(&ipf->n4frag_cnt[IPF_NFRAGS_TOO_SMALL],
                         &ipf_status->v4.nfrag_too_small);
@@ -1461,7 +1473,7 @@ ipf_get_status(struct ipf *ipf, struct ipf_status *ipf_status)
                         &ipf_status->v6.nfrag_accepted);
     atomic_read_relaxed(&ipf->n6frag_cnt[IPF_NFRAGS_COMPL_SENT],
                         &ipf_status->v6.nfrag_completed_sent);
-    atomic_read_relaxed(&ipf->n6frag_cnt[IPF_NFRAGS_EXPD_SENT],
+    atomic_read_relaxed(&ipf->n6frag_cnt[IPF_NFRAGS_EXPIRED],
                         &ipf_status->v6.nfrag_expired_sent);
     atomic_read_relaxed(&ipf->n6frag_cnt[IPF_NFRAGS_TOO_SMALL],
                         &ipf_status->v6.nfrag_too_small);
