@@ -33,6 +33,7 @@
 #include <sys/time.h>
 
 #include "byte-order.h"
+#include "coverage.h"
 #include "csum.h"
 #include "dp-packet.h"
 #include "netdev.h"
@@ -49,6 +50,11 @@
 VLOG_DEFINE_THIS_MODULE(native_tnl);
 static struct vlog_rate_limit err_rl = VLOG_RATE_LIMIT_INIT(60, 5);
 
+COVERAGE_DEFINE(native_tnl_l3csum_checked);
+COVERAGE_DEFINE(native_tnl_l3csum_err);
+COVERAGE_DEFINE(native_tnl_l4csum_checked);
+COVERAGE_DEFINE(native_tnl_l4csum_err);
+
 #define VXLAN_HLEN   (sizeof(struct udp_header) +         \
                       sizeof(struct vxlanhdr))
 
@@ -61,8 +67,29 @@ static struct vlog_rate_limit err_rl = VLOG_RATE_LIMIT_INIT(60, 5);
 uint16_t tnl_udp_port_min = 32768;
 uint16_t tnl_udp_port_max = 61000;
 
-void *
-netdev_tnl_ip_extract_tnl_md(struct dp_packet *packet, struct flow_tnl *tnl,
+ovs_be16
+netdev_tnl_get_src_port(struct dp_packet *packet)
+{
+    uint32_t hash;
+
+    if (OVS_LIKELY(dp_packet_rss_valid(packet))) {
+        hash = dp_packet_get_rss_hash(packet);
+    } else {
+        struct flow flow;
+
+        flow_extract(packet, &flow);
+        hash = flow_hash_5tuple(&flow, 0);
+
+        dp_packet_set_rss_hash(packet, hash);
+    }
+
+    hash = ((uint64_t) hash * (tnl_udp_port_max - tnl_udp_port_min)) >> 32;
+
+    return htons(hash + tnl_udp_port_min);
+}
+
+static void *
+ip_extract_tnl_md(struct dp_packet *packet, struct flow_tnl *tnl,
                   unsigned int *hlen)
 {
     void *nh;
@@ -86,16 +113,20 @@ netdev_tnl_ip_extract_tnl_md(struct dp_packet *packet, struct flow_tnl *tnl,
               ((char *)nh - (char *)dp_packet_data(packet));
 
     if (IP_VER(ip->ip_ihl_ver) == 4) {
-
+        bool bad_csum = dp_packet_ip_checksum_bad(packet);
         ovs_be32 ip_src, ip_dst;
 
         /* A packet coming from a network device might have the
          * csum already checked. In this case, skip the check. */
-        if (OVS_UNLIKELY(!dp_packet_hwol_l3_csum_ipv4_ol(packet))) {
-            if (csum(ip, IP_IHL(ip->ip_ihl_ver) * 4)) {
-                VLOG_WARN_RL(&err_rl, "ip packet has invalid checksum");
-                return NULL;
-            }
+        if (OVS_UNLIKELY(!bad_csum
+                         && !dp_packet_hwol_l3_csum_ipv4_ol(packet))) {
+            COVERAGE_INC(native_tnl_l3csum_checked);
+            bad_csum = csum(ip, IP_IHL(ip->ip_ihl_ver) * 4);
+        }
+        if (OVS_UNLIKELY(bad_csum)) {
+            COVERAGE_INC(native_tnl_l3csum_err);
+            VLOG_WARN_RL(&err_rl, "ip packet has invalid checksum");
+            return NULL;
         }
 
         if (ntohs(ip->ip_tot_len) > l3_size) {
@@ -208,15 +239,18 @@ udp_extract_tnl_md(struct dp_packet *packet, struct flow_tnl *tnl,
 {
     struct udp_header *udp;
 
-    udp = netdev_tnl_ip_extract_tnl_md(packet, tnl, hlen);
+    udp = ip_extract_tnl_md(packet, tnl, hlen);
     if (!udp) {
         return NULL;
     }
 
     if (udp->udp_csum) {
-        if (OVS_LIKELY(!dp_packet_ol_l4_csum_partial(packet)) &&
-            OVS_UNLIKELY(!dp_packet_l4_checksum_good(packet))) {
+        bool bad_csum = dp_packet_l4_checksum_bad(packet);
+
+        if (OVS_LIKELY(!bad_csum && !dp_packet_ol_l4_csum_partial(packet))
+            && OVS_UNLIKELY(!dp_packet_l4_checksum_good(packet))) {
             uint32_t csum;
+            COVERAGE_INC(native_tnl_l4csum_checked);
             if (netdev_tnl_is_header_ipv6(dp_packet_data(packet))) {
                 csum = packet_csum_pseudoheader6(dp_packet_l3(packet));
             } else {
@@ -227,9 +261,11 @@ udp_extract_tnl_md(struct dp_packet *packet, struct flow_tnl *tnl,
                                  ((const unsigned char *)udp -
                                   (const unsigned char *)dp_packet_eth(packet)
                                  ));
-            if (csum_finish(csum)) {
-                return NULL;
-            }
+            bad_csum = csum_finish(csum);
+        }
+        if (OVS_UNLIKELY(bad_csum)) {
+            COVERAGE_INC(native_tnl_l4csum_err);
+            return NULL;
         }
         tnl->flags |= FLOW_TNL_F_CSUM;
     }
@@ -254,7 +290,7 @@ dp_packet_tnl_ol_process(struct dp_packet *packet,
 
             if (IP_VER(ip->ip_ihl_ver) == 4) {
                 dp_packet_hwol_set_tx_ipv4(packet);
-                dp_packet_hwol_tx_ip_csum(packet);
+                dp_packet_hwol_set_tx_ip_csum(packet);
             } else if (IP_VER(ip->ip_ihl_ver) == 6) {
                 dp_packet_hwol_set_tx_ipv6(packet);
             }
@@ -276,14 +312,18 @@ netdev_tnl_push_udp_header(const struct netdev *netdev OVS_UNUSED,
     uint16_t l3_ofs = packet->l3_ofs;
     uint16_t l4_ofs = packet->l4_ofs;
     struct udp_header *udp;
+    ovs_be16 udp_src;
     int ip_tot_size;
 
+    /* We may need to re-calculate the hash and this has to be done before
+     * modifying the packet. */
+    udp_src = netdev_tnl_get_src_port(packet);
+
     dp_packet_tnl_ol_process(packet, data);
     udp = netdev_tnl_push_ip_header(packet, data->header, data->header_len,
                                     &ip_tot_size, 0);
 
-    /* set udp src port */
-    udp->udp_src = netdev_tnl_get_src_port(packet);
+    udp->udp_src = udp_src;
     udp->udp_len = htons(ip_tot_size);
 
     if (udp->udp_csum) {
@@ -421,7 +461,7 @@ parse_gre_header(struct dp_packet *packet,
     unsigned int ulen;
     uint16_t greh_protocol;
 
-    greh = netdev_tnl_ip_extract_tnl_md(packet, tnl, &ulen);
+    greh = ip_extract_tnl_md(packet, tnl, &ulen);
     if (!greh) {
         return -EINVAL;
     }
@@ -602,7 +642,7 @@ netdev_erspan_pop_header(struct dp_packet *packet)
         goto err;
     }
 
-    greh = netdev_tnl_ip_extract_tnl_md(packet, tnl, &ulen);
+    greh = ip_extract_tnl_md(packet, tnl, &ulen);
     if (!greh) {
         goto err;
     }
@@ -831,13 +871,18 @@ netdev_gtpu_push_header(const struct netdev *netdev,
     struct netdev_vport *dev = netdev_vport_cast(netdev);
     struct udp_header *udp;
     struct gtpuhdr *gtpuh;
+    ovs_be16 udp_src;
     int ip_tot_size;
     unsigned int payload_len;
 
+    /* We may need to re-calculate the hash and this has to be done before
+     * modifying the packet. */
+    udp_src = netdev_tnl_get_src_port(packet);
+
     payload_len = dp_packet_size(packet);
     udp = netdev_tnl_push_ip_header(packet, data->header, data->header_len,
                                     &ip_tot_size, 0);
-    udp->udp_src = netdev_tnl_get_src_port(packet);
+    udp->udp_src = udp_src;
     udp->udp_len = htons(ip_tot_size);
     /* Postpone checksum to the egress netdev. */
     dp_packet_hwol_set_csum_udp(packet);
@@ -1027,7 +1072,7 @@ netdev_srv6_pop_header(struct dp_packet *packet)
     }
 
     pkt_metadata_init_tnl(md);
-    if (!netdev_tnl_ip_extract_tnl_md(packet, tnl, &hlen)) {
+    if (!ip_extract_tnl_md(packet, tnl, &hlen)) {
         goto err;
     }
 
