@@ -441,6 +441,26 @@ def idl_set(idl, commands, step):
 
             s = txn.insert(idl.tables["simple"])
             s.i = int(args[0])
+        elif name == "insert_no_columns_changed":
+            txn.insert(idl.tables["simple"])
+        elif name == "insert_uuid":
+            if len(args) != 2:
+                sys.stderr.write('"set" command requires 2 argument\n')
+                sys.exit(1)
+
+            s = txn.insert(idl.tables["simple"], new_uuid=uuid.UUID(args[0]),
+                           persist_uuid=True)
+            s.i = int(args[1])
+        elif name == "insert_uuid_uref":
+            if len(args) != 2:
+                sys.stderr.write('"set" command requires 2 argument\n')
+                sys.exit(1)
+
+            s4 = txn.insert(idl.tables["simple4"], new_uuid=uuid.UUID(args[1]),
+                            persist_uuid=True)
+            s3 = txn.insert(idl.tables["simple3"], new_uuid=uuid.UUID(args[0]),
+                            persist_uuid=True)
+            s3.uref = s4
         elif name == "delete":
             if len(args) != 1:
                 sys.stderr.write('"delete" command requires 1 argument\n')
@@ -503,7 +523,7 @@ def idl_set(idl, commands, step):
             print("%03d: destroy" % step)
             sys.stdout.flush()
             txn.abort()
-            return
+            return True
         elif name == "linktest":
             l1_0 = txn.insert(idl.tables["link1"])
             l1_0.i = 1
@@ -627,6 +647,8 @@ def idl_set(idl, commands, step):
     sys.stdout.write("\n")
     sys.stdout.flush()
 
+    return status != ovs.db.idl.Transaction.ERROR
+
 
 def update_condition(idl, commands, step):
     next_cond_seqno = 0
@@ -791,7 +813,13 @@ def do_idl(schema_file, remote, *commands):
             next_cond_seqno = update_condition(idl, command, step)
             step += 1
         elif not command.startswith("["):
-            idl_set(idl, command, step)
+            if not idl_set(idl, command, step):
+                # If idl_set() returns false, then no transaction
+                # was sent to the server and most likely seqno
+                # would remain the same.  And the above 'Wait for update'
+                # for loop poller.block() would never return.
+                # So set seqno to 0.
+                seqno = 0
             step += 1
         else:
             json = ovs.json.from_string(command)
